﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using System.IO;
using Microsoft.Kinect;

namespace MultiKinect
{
    /// <summary>
    /// The main window - Handles all Init and Main loop - Kinect Collection
    /// </summary>
    public partial class MainWindow : Window
    {
        private KinectSensorCollection mkSensors; //all sensors
        //Skeleton Data Containers
        private List<Skeleton[]> allSkeletons; //all skeletons from all kinects
        private List<Skeleton> unsafeList = new List<Skeleton>();
        private List<MappingPanel> mappings = new List<MappingPanel>();
        private List<UserRadar> radars = new List<UserRadar>();
        //tracking
        private MultiSkeleton user1 = new MultiSkeleton();
        private MultiSkeleton user2 = new MultiSkeleton();
        //settings
        private int tiltValue = 0;
        //NET
        private OscFeedback oscMain;
        //tracking mode
        private bool trackingSeated = false;

        //Data containers for the RGB Image
        private List<WriteableBitmap> rgbOutputs;
        private byte[] colorPixels;
        private List<Image> rgbImages;


        public MainWindow()
        {
            InitializeComponent();
        }

        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            allSkeletons = new List<Skeleton[]>();
            oscMain = new OscFeedback();
            wrapPanelOSCfeedback.Children.Add(oscMain);
            textBoxOscIp.Text = System.Net.IPAddress.Loopback.ToString();
            rgbOutputs = new List<WriteableBitmap>();
            rgbImages = new List<Image>();
            user1.Name = "User1";
            user2.Name = "User2";
            //kinect initalization
            this.mkSensors = KinectSensor.KinectSensors;
            int kCount = 0;
            foreach (KinectSensor sensor in mkSensors)
            {
                textBox1.AppendText("\n Kinect( " + sensor.UniqueKinectId.ToString() + " ): " + sensor.Status.ToString());
                if (sensor.Status == KinectStatus.Connected)
                {
                   
                    //Turn On Color Stream
                    sensor.ColorStream.Enable(ColorImageFormat.RgbResolution640x480Fps30);

                    //Allocate Space for pixel arrays
                    this.colorPixels = new byte[sensor.ColorStream.FramePixelDataLength];

                    //Add writeable Bitmap for display of RGB stream
                    this.rgbOutputs.Add(new WriteableBitmap(sensor.ColorStream.FrameWidth, sensor.ColorStream.FrameHeight, 96.0, 96.0, PixelFormats.Bgr32, null));

                    //Adding an image for output of RGB
                    Image outImage = new Image();
                    outImage.Name = "rgbImage" + kCount.ToString();
                    outImage.Width = 160;
                    outImage.Height = 120;
                    rgbImages.Add(outImage);
                    outImage.Source = this.rgbOutputs[kCount];
                    wrapPanel1.Children.Add(outImage);

                    //event handling for rgb data
                    sensor.ColorFrameReady += this.SensorColorFrameReady;
             
                    //MAPPING CONTROLS
                    MappingPanel mapView = new MappingPanel();
                    wrapPanelMapping.Children.Add(mapView);
                    mappings.Add(mapView);

                    UserRadar radar = new UserRadar();
                    grid1.Children.Add(radar);
                    radars.Add(radar);
                    radar.Margin = new Thickness( (double)((radars.Count-1) * 300.0) , 0, 0, 0);

                    sensor.SkeletonStream.Enable();
                    //sensor.SkeletonStream.AppChoosesSkeletons = true;
                    //Allocate space for Skeleton tracking
                    allSkeletons.Add(new Skeleton[sensor.SkeletonStream.FrameSkeletonArrayLength]);
                    //event handling for skeleton
                    sensor.SkeletonFrameReady += this.SensorSkeletonFrameReady;
                    //frame event handling
                    CompositionTargetEx.FrameUpdating += new EventHandler<RenderingEventArgs>(CompositionTargetEx_FrameUpdating);

                    try
                    {
                        sensor.Start();
                        mapView.CamActive(true);
                    }
                    catch (IOException)
                    {
                        int sensorNumber = mkSensors.IndexOf(sensor);
                        textBox1.AppendText("\n Failed to start device ID: " + mkSensors[sensorNumber].UniqueKinectId.ToString() + " internal number " + sensorNumber.ToString());
                        mkSensors[sensorNumber].Dispose();
                        mapView.CamActive(false);
                    }
                }
                kCount++;
            }
        }

        //Update Event
        private void CompositionTargetEx_FrameUpdating(object sender, RenderingEventArgs e)
        {
            textBox1.Text = "";
            if (user1.Skeleton != null && user1.GlobalPosition != null)
            {
                textBox1.AppendText("user1: " + user1.Skeleton.Position.X.ToString() +
                                   " " + user1.Skeleton.Position.Y.ToString() +
                                   " " + user1.Skeleton.Position.Z.ToString() + " ----- ID: " + user1.Skeleton.TrackingId.ToString());
                textBox1.AppendText("\n" + "global: " + user1.GlobalPosition.X.ToString() +
                                   " " + user1.GlobalPosition.Y.ToString() +
                                   " " + user1.GlobalPosition.Z.ToString());
                textBox1.AppendText("\n height: " + user1.Skeleton.Height().ToString());
            }
            if (user2.Skeleton != null && user2.GlobalPosition != null)
            {
                textBox1.AppendText("\n" + "user2: " + user2.Skeleton.Position.X.ToString() +
                                   " " + user2.Skeleton.Position.Y.ToString() +
                                   " " + user2.Skeleton.Position.Z.ToString());
                textBox1.AppendText("\n" + "global: " + user2.GlobalPosition.X.ToString() +
                                   " " + user2.GlobalPosition.Y.ToString() +
                                   " " + user2.GlobalPosition.Z.ToString() + " ----- ID: " + user2.Skeleton.TrackingId.ToString());
                textBox1.AppendText("\n height: " + user2.Skeleton.Height().ToString());
            }

            if (oscMain.IsStarted())
            {
                if(user1.TrackingState == SkeletonTrackingState.Tracked)
                    oscMain.sendMessage(user1);
            }
        }

        private void Window_Closed(object sender, EventArgs e)
        {
            foreach (KinectSensor sensor in mkSensors)
            {
                if (sensor.IsRunning)
                {
                    sensor.Stop();
                }
            }
        }

        private void SensorColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame colorFrame = e.OpenColorImageFrame())
            {
                if (colorFrame != null)
                { 
                    int sensorNumber = mkSensors.IndexOf(sender as KinectSensor);
                    colorFrame.CopyPixelDataTo(this.colorPixels);

                    //write to the output
                    if (this.rgbOutputs[sensorNumber] != null)
                    {
                        this.rgbOutputs[sensorNumber].WritePixels(
                            new Int32Rect(0, 0, this.rgbOutputs[sensorNumber].PixelWidth, this.rgbOutputs[sensorNumber].PixelHeight),
                            this.colorPixels,
                            this.rgbOutputs[sensorNumber].PixelWidth * sizeof(int),
                            0);
                    }
                    else
                    {
                        this.textBox1.Text = "Failed to write Pixel Data - no Bitmap Data defined";
                        return;
                    }
                }
            }
        }

        private void SensorSkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            int sensorNumber = mkSensors.IndexOf(sender as KinectSensor);

            using (SkeletonFrame skeletonFrame = e.OpenSkeletonFrame())
            {
                if (skeletonFrame != null && this.allSkeletons[sensorNumber] != null)
                {
                    skeletonFrame.CopySkeletonDataTo(this.allSkeletons[sensorNumber]);
                }
            }

            //Remove Unused non existing IDs
            user1 = RemoveInactiveSkeletons(user1);
            user2 = RemoveInactiveSkeletons(user2);
            RemoveInaktiveRadarUsers(sensorNumber);
            //MAIN tracking logic
            SimpleMultiTracking(sensorNumber);
            
        }

        private void SimpleMultiTracking(int sensorNumber)
        {
            foreach (Skeleton skeleton in allSkeletons[sensorNumber])
            {
                if (skeleton != null)
                {
                    if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                    {
                        //Entry 1) when the user1 is not tracked - we assign the skeleton 
                        //(but only if its not assigned to the any user) 
                        if (user1.TrackingState != SkeletonTrackingState.Tracked
                            && !user1.assignedSkeletons.ContainsKey(skeleton.TrackingId)
                            && !user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                        {
                            user1.Skeleton = skeleton;
                            user1.TrackingState = skeleton.TrackingState;
                            user1.KinectNumber = sensorNumber;
                            user1.assignedSkeletons.Add(skeleton.TrackingId, skeleton);
                        }
                        
                        //2) when the user1 is tracked and the skeleton is from another kinect
                        // and not on the list from both users we compare it with the active user 1 skeleton
                        if (user1.TrackingState == SkeletonTrackingState.Tracked 
                            && user1.KinectNumber != sensorNumber
                            && !user1.assignedSkeletons.ContainsKey(skeleton.TrackingId)
                            && !user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                        { 
                            //compare and if same throw back which one is better
                            if (CompareSkeletons(user1.Skeleton, user1.KinectNumber, skeleton, sensorNumber))
                            {
                                user1.Skeleton = BetterSkeleton(user1.Skeleton, skeleton);
                                //if the second skeleton was better update sensor number
                                if (user1.Skeleton.TrackingId == skeleton.TrackingId)
                                    user1.KinectNumber = sensorNumber;

                                //in any case add the skeleton to the users list
                                user1.assignedSkeletons.Add(skeleton.TrackingId, skeleton);
                            }
                            
                        }

                        //same as above but for the second user
                        if (user2.TrackingState != SkeletonTrackingState.Tracked 
                            && !user1.assignedSkeletons.ContainsKey(skeleton.TrackingId)
                            && !user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                        {
                            user2.Skeleton = skeleton;
                            user2.TrackingState = skeleton.TrackingState;
                            user2.KinectNumber = sensorNumber;
                            user2.assignedSkeletons.Add(skeleton.TrackingId, skeleton);
                        }

                        //compare logic for second user
                        if (user2.TrackingState == SkeletonTrackingState.Tracked
                            && user2.KinectNumber != sensorNumber
                            && !user1.assignedSkeletons.ContainsKey(skeleton.TrackingId)
                            && !user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                        {
                            //compare and if same throw back which one is better
                            if (CompareSkeletons(user2.Skeleton, user2.KinectNumber, skeleton, sensorNumber))
                            {
                                user2.Skeleton = BetterSkeleton(user2.Skeleton, skeleton);
                                //if the second skeleton was better update sensor number
                                if (user2.Skeleton.TrackingId == skeleton.TrackingId)
                                    user2.KinectNumber = sensorNumber;

                                //in any case add the skeleton to the users list
                                user2.assignedSkeletons.Add(skeleton.TrackingId, skeleton);
                            }

                        }


                        //update skeletons if ID matches
                        if (skeleton.TrackingId == user1.Skeleton.TrackingId)
                        {
                            user1.TrackingState = skeleton.TrackingState;
                            user1.Skeleton = skeleton;
                            user1.GlobalPosition = ConvertSkeletonPointToGlobal(sensorNumber, skeleton.Position);
                            user1.KinectNumber = sensorNumber;
                        }

                        if (skeleton.TrackingId == user2.Skeleton.TrackingId 
                            && user2.Skeleton.TrackingId != user1.Skeleton.TrackingId)
                        {
                            user2.TrackingState = skeleton.TrackingState;
                            user2.Skeleton = skeleton;
                            user2.GlobalPosition = ConvertSkeletonPointToGlobal(sensorNumber, skeleton.Position);
                            user2.KinectNumber = sensorNumber;
                        }
                    }

                    //Radar Display
                    if (skeleton.TrackingState != SkeletonTrackingState.NotTracked)
                    {
                        //User Radar Output
                        if (radars[sensorNumber].PlayerExists(skeleton.TrackingId))
                        {
                            //update position
                            Point posi = MapToScreenPoint(sensorNumber, skeleton.Position);
                            
                            if (skeleton.TrackingState == SkeletonTrackingState.Tracked)
                            {
                                if (user1.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                                    radars[sensorNumber].UpdatePosition(posi, skeleton.TrackingId, "TU1");
                                else if (user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                                    radars[sensorNumber].UpdatePosition(posi, skeleton.TrackingId, "TU2");
                                else
                                    radars[sensorNumber].UpdatePosition(posi, skeleton.TrackingId, "T");
                            }
                            if (skeleton.TrackingState == SkeletonTrackingState.PositionOnly)
                                radars[sensorNumber].UpdatePosition(posi, skeleton.TrackingId, "P");
                        }
                        else
                        {
                            //add to radar list
                            Point posi = MapToScreenPoint(sensorNumber, skeleton.Position);
                            if(skeleton.TrackingState == SkeletonTrackingState.Tracked)
                            {
                                if(user1.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                                    radars[sensorNumber].AddPlayer(posi, skeleton.TrackingId, "TU1");
                                else if (user2.assignedSkeletons.ContainsKey(skeleton.TrackingId))
                                    radars[sensorNumber].AddPlayer(posi, skeleton.TrackingId, "TU2");
                                else
                                    radars[sensorNumber].AddPlayer(posi, skeleton.TrackingId, "T");
                            }
                            if(skeleton.TrackingState == SkeletonTrackingState.PositionOnly)
                                radars[sensorNumber].AddPlayer(posi, skeleton.TrackingId, "P");
                        }
                    }
                }
            }
        }

        private MultiSkeleton RemoveInactiveSkeletons(MultiSkeleton user)
        {
            List<int> removeIDs = new List<int>();
            foreach (int id in user.assignedSkeletons.Keys)
            {
                bool doesHave = false;
                for (int y = 0; y < allSkeletons.Count; y++)
                {
                    for (int i = 0; i < allSkeletons[y].Length; i++)
                    {
                        if (id == allSkeletons[y][i].TrackingId)
                            doesHave = true;
                    }
                }
                if (!doesHave)
                    removeIDs.Add(id);
            }

            if (removeIDs.Count > 0)
            {
                foreach (int id in removeIDs)
                {
                    try
                    {
                        user.assignedSkeletons.Remove(id);
                    }
                    catch (Exception)
                    { 
                    
                    }
                }
            }

            //when the Main Tracked id was removed add the first from the assigned list (if any)
            if (removeIDs.Contains(user.Skeleton.TrackingId) && user.assignedSkeletons.Count > 0)
            {
                user.Skeleton = user.assignedSkeletons.Values.First();
                for (int x = 0; x < allSkeletons.Count; ++x)
                {
                    if (allSkeletons[x].Contains(user.Skeleton))
                    {
                        user.KinectNumber = x;
                        break;
                    }
                }
            }
                
            //if all where lost
            if (user.assignedSkeletons.Count == 0)
            {
                user.TrackingState = SkeletonTrackingState.NotTracked;
                user.KinectNumber = -1;
            }

            return user;
        }

        private void RemoveInaktiveRadarUsers(int sensorNumber)
        {
            List<int> ids = radars[sensorNumber].getIDs();
            for (int x = 0; x < ids.Count; ++x)
            {
                bool doesHave = false;
                for (int i = 0; i < allSkeletons[sensorNumber].Length; i++)
                {
                    if (ids[x] == allSkeletons[sensorNumber][i].TrackingId)
                        doesHave = true;
                }
                if (!doesHave)
                    radars[sensorNumber].RemovePlayer(ids[x]);
            }        
        }

        private Skeleton BetterSkeleton(Skeleton skeleton1, Skeleton skeleton2)
        {
            int score1 = 0;
            int score2 = 0;

            foreach (Joint ji in skeleton1.Joints)
            {
                if (ji.TrackingState == JointTrackingState.Tracked)
                    score1++;
            }

            foreach (Joint ji in skeleton2.Joints)
            {
                if (ji.TrackingState == JointTrackingState.Tracked)
                    score2++;
            }

            if (score1 >= score2)
                return skeleton1;

            if (score2 > score1)
                return skeleton2;

            return skeleton1;
        }

        private bool CompareSkeletons(Skeleton skeleton1, int sensor1, Skeleton skeleton2, int sensor2)
        {
            SkeletonPoint point1 = ConvertSkeletonPointToGlobal(sensor1, skeleton1.Position);
            SkeletonPoint point2 = ConvertSkeletonPointToGlobal(sensor2, skeleton2.Position);

            bool comparer = false;
            //!TODO: improve location comparison (e.g. X, Y, Z depency) 
            if (CompareSkeletonPosition(point1, point2, 0.15f))
            {
                if (!trackingSeated)
                {
                    double height1 = skeleton1.Height();
                    double height2 = skeleton2.Height();
                    if (Math.Abs((height1 - height2)) < 0.05f)
                    { 
                        //see if the users height matches to further improve accuracy
                        comparer = true;
                    }
                }
                else
                {
                    comparer = true;
                }
            }

            return comparer;
        }

        private bool CompareSkeletonPosition(SkeletonPoint p1, SkeletonPoint p2, float threshold)
        {
            //in this simple approach we only compare the Z position
            if (Math.Abs(p1.Z - p2.Z) < threshold) //&& Math.Abs(p1.X - p2.X) < 0.5f
                return true;

            return false;
        }

        public SkeletonPoint ConvertSkeletonPointToGlobal(int sensorNumber, SkeletonPoint position)
        {
            SkeletonPoint b4temp = position;
            MultiKinectHelpers.setTransform(mappings[sensorNumber].Offset);
            position = MultiKinectHelpers.normalizeKinectPoint(position);
            return position;
        }

        public Point MapToScreenPoint(int sensorNumber, SkeletonPoint position)
        {
            //DepthImagePoint depthPoint = mkSensors[sensorNumber].CoordinateMapper.MapSkeletonPointToDepthPoint(position, DepthImageFormat.Resolution640x480Fps30);
            Point p = new Point();
            p.X = mapDouble(position.X, -2.0, 2.0, 0.0, 300.0);
            p.Y = mapDouble(position.Z, 0.0, 3.8, 0.0, 200.0);
            return p;
        }


        private double mapDouble(double input, double inputMin, double inputMax, double outputMin, double outputMax)
        {
            return outputMin + (input - inputMin) * (outputMax - outputMin) / (inputMax - inputMin);
        }
  
        private void setIpButton_Click(object sender, RoutedEventArgs e)
        {
            bool temp;
            temp = oscMain.setIpAddress(textBoxOscIp.Text);
            temp = oscMain.setPort(textBoxOscPort.Text);
            if (!temp)
                textBox1.Text = "Failed to set OSC IP - Check input";

            oscMain.AddressChanged();
        }

        private void checkBoxOSCsend_Click(object sender, RoutedEventArgs e)
        {
            if (checkBoxOSCsend.IsChecked.Value)
            {
                oscMain.startOsc();
            }
            else
            {
                oscMain.stopOsc();
            }
        }

        private void buttonSetTilt_Click(object sender, RoutedEventArgs e)
        {
            int sliderValue = (int)Math.Round(sliderTilt.Value);
            labelTiltOut.Content = sliderValue.ToString();
            
            foreach (KinectSensor sensor in mkSensors)
            {
                if (sensor != null && sensor.IsRunning)
                {
                    tiltValue = sliderValue;
                    sensor.ElevationAngle = sliderValue;
                }
            }
        }

        private void checkBoxSeatedMode_Click(object sender, RoutedEventArgs e)
        {
            foreach (KinectSensor sensor in mkSensors)
            {
                if (checkBoxSeatedMode.IsChecked.Value)
                {
                    sensor.SkeletonStream.TrackingMode = SkeletonTrackingMode.Seated;
                    trackingSeated = true;
                }
                else
                {
                    sensor.SkeletonStream.TrackingMode = SkeletonTrackingMode.Default;
                    trackingSeated = false;
                }
            }
        }
        
    }

    public static class CompositionTargetEx
    {
        private static TimeSpan _last = TimeSpan.Zero;
        private static event EventHandler<RenderingEventArgs> _FrameUpdating;
        
        public static event EventHandler<RenderingEventArgs> FrameUpdating
        {
            add
            {
                if (_FrameUpdating == null)
                    CompositionTarget.Rendering += CompositionTarget_Rendering; 
                _FrameUpdating += value;
            }
            remove
            {
                _FrameUpdating -= value; 
                if (_FrameUpdating == null)
                    CompositionTarget.Rendering -= CompositionTarget_Rendering;
            }
        }

        static void CompositionTarget_Rendering(object sender, EventArgs e)
        {
            RenderingEventArgs args = (RenderingEventArgs)e;
            if (args.RenderingTime == _last) 
                return;
            _last = args.RenderingTime; 
            _FrameUpdating(sender, args);
        }
    }

    public class MultiSkeleton
    {
        private Skeleton _skeleton; //the currently active skeleton
        private int kinectNumber;
        private SkeletonTrackingState trackingState;
        private SkeletonPoint globalPosition;
        private string name;
        public Dictionary<int, Skeleton> assignedSkeletons; //a list of skeleton ids which are the same user

        public MultiSkeleton()
        {
            _skeleton = new Skeleton();
            _skeleton.TrackingId = -1;
            kinectNumber = -1;
            assignedSkeletons = new Dictionary<int, Skeleton>();
        }

        public Skeleton Skeleton
        { 
            set { this._skeleton = value; }
            get { return this._skeleton; }
        }

        public int KinectNumber
        {
            set { this.kinectNumber = value; }
            get { return this.kinectNumber; }
        }

        public SkeletonTrackingState TrackingState
        {
            set { this.trackingState = value; }
            get { return this.trackingState; }
        }

        public SkeletonPoint GlobalPosition
        {
            set { this.globalPosition = value; }
            get { return this.globalPosition; }
        }

        public string Name
        {
            set { this.name = value; }
            get { return this.name; }
        }
    }
}
